import sys
import threading
import numpy as np
from sentence_transformers import SentenceTransformer
from typing import List, Optional


class Embedder:
    def __init__(self, model_name: str = "sentence-transformers/all-MiniLM-L6-v2"):
        print(f"[DEBUG] Embedder.__init__ starting (background loading enabled)", file=sys.stderr, flush=True)
        self.model_name = model_name
        self._model: Optional[SentenceTransformer] = None
        self._dim: Optional[int] = None
        self._loaded = threading.Event()

        def _load_model():
            print(f"[DEBUG] Loading sentence-transformers model: {self.model_name}", file=sys.stderr, flush=True)
            self._model = SentenceTransformer(self.model_name)
            self._dim = self._model.get_sentence_embedding_dimension()
            self._loaded.set()
            print(f"[DEBUG] Model loaded successfully, dim={self._dim}", file=sys.stderr, flush=True)

        threading.Thread(target=_load_model, daemon=True).start()
        print(f"[DEBUG] Embedder.__init__ completed (model loading started in background)", file=sys.stderr, flush=True)
    
    def _ensure_model_loaded(self):
        """Ensure the model is loaded, waiting if necessary."""
        self._loaded.wait()
    
    @property
    def model(self) -> SentenceTransformer:
        """Get the model, loading it if necessary."""
        self._ensure_model_loaded()
        assert self._model is not None
        return self._model

    @property
    def dim(self) -> int:
        """Get the embedding dimension, loading model if necessary."""
        self._ensure_model_loaded()
        assert self._dim is not None
        return self._dim

    def encode_batch(self, texts: List[str], batch_size: int = 64) -> np.ndarray:
        """Encode texts in batches, return L2 normalized embeddings."""
        self._ensure_model_loaded()
        embeddings = []
        for i in range(0, len(texts), batch_size):
            batch = texts[i:i + batch_size]
            emb = self._model.encode(batch, convert_to_numpy=True, normalize_embeddings=True)
            embeddings.append(emb)
        return np.vstack(embeddings) if embeddings else np.empty((0, self._dim))

    def encode_single(self, text: str) -> np.ndarray:
        """Encode single text, L2 normalized."""
        self._ensure_model_loaded()
        return self._model.encode([text], convert_to_numpy=True, normalize_embeddings=True)[0]


import re
from pathlib import Path
from typing import List, Tuple, Optional
from pypdf import PdfReader
from docx import Document
from charset_normalizer import detect
from .utils import compute_checksum, normalize_text


def read_file(path: Path) -> Optional[str]:
    """Read file content based on extension."""
    try:
        if path.suffix.lower() in ['.txt', '.md']:
            return read_text_file(path)
        elif path.suffix.lower() == '.pdf':
            return read_pdf_file(path)
        elif path.suffix.lower() == '.docx':
            return read_docx_file(path)
        else:
            return None
    except Exception as e:
        print(f"Error reading {path}: {e}", file=__import__('sys').stderr)
        return None


def read_text_file(path: Path) -> str:
    """Read text file with charset detection."""
    with open(path, 'rb') as f:
        raw = f.read()
    detected = detect(raw)
    encoding = detected.get('encoding', 'utf-8')
    return raw.decode(encoding, errors='replace')


def read_pdf_file(path: Path) -> str:
    """Extract text from PDF."""
    reader = PdfReader(path)
    text = ""
    for page in reader.pages:
        text += page.extract_text() + "\n"
    return text.strip()


def read_docx_file(path: Path) -> str:
    """Extract text from DOCX."""
    doc = Document(path)
    text = ""
    for para in doc.paragraphs:
        text += para.text + "\n\n"
    return text.strip()


def chunk_text(text: str, chunk_size: int = 1000, overlap: int = 200) -> List[Tuple[str, int, int]]:
    """Chunk text into overlapping segments, preserving paragraphs."""
    chunks = []
    start = 0
    while start < len(text):
        end = start + chunk_size
        if end < len(text):
            # Try to break at paragraph boundary
            para_end = text.rfind('\n\n', start, end)
            if para_end > start + chunk_size // 2:
                end = para_end + 2
        chunk = text[start:end].strip()
        if chunk:
            chunks.append((chunk, start, end))
        start = end - overlap
        if start >= len(text):
            break
    return chunks


def process_file(path: Path) -> List[Tuple[str, int, int, str]]:
    """Process file: read, chunk, return (text, start, end, checksum)."""
    content = read_file(path)
    if content is None:
        return []
    checksum = compute_checksum(content)
    chunks = chunk_text(content)
    return [(text, start, end, checksum) for text, start, end in chunks]


import sys
from typing import Optional
from fastmcp import FastMCP
from .types import Config, ReindexResult, SearchResult
from .embedder import Embedder
from .index import IndexManager
import asyncio


class MicroRAGServer:
    def __init__(self, config: Config, name: str = "micro-rag-mcp", description: Optional[str] = None):
        print("[DEBUG] MicroRAGServer.__init__ starting...", file=sys.stderr, flush=True)
        self.config = config
        print("[DEBUG] Creating Embedder...", file=sys.stderr, flush=True)
        self.embedder = Embedder()
        print("[DEBUG] Creating IndexManager...", file=sys.stderr, flush=True)
        self.index_manager = IndexManager(config, self.embedder)
        print("[DEBUG] Creating FastMCP instance...", file=sys.stderr, flush=True)
        self.mcp = FastMCP(name=name, instructions=description)
        print("[DEBUG] Registering tools...", file=sys.stderr, flush=True)

        @self.mcp.tool()
        async def search(query: str, top_k: int = 5, score_threshold: float = 0.2) -> list[SearchResult]:
            """Search the document index for relevant chunks."""
            results = self.index_manager.search(query, top_k, score_threshold)
            return [SearchResult(**r) for r in results]

        @self.mcp.tool()
        async def reindex(force: bool = False, path_glob: Optional[str] = None) -> ReindexResult:
            """Reindex documents incrementally."""
            result = self.index_manager.reindex(force, path_glob)
            return ReindexResult(**result)
        
        print("[DEBUG] Tools registered successfully", file=sys.stderr, flush=True)

    def run(self):
        print("[DEBUG] Starting MCP server run()...", file=sys.stderr, flush=True)
        self.mcp.run()


import hashlib
import os
import re
from pathlib import Path
from typing import List, Tuple


def compute_checksum(text: str) -> str:
    """Compute SHA256 checksum of normalized text."""
    normalized = normalize_text(text)
    return f"sha256:{hashlib.sha256(normalized.encode('utf-8')).hexdigest()}"


def normalize_text(text: str) -> str:
    """Normalize text for consistent checksums: strip, normalize newlines."""
    return re.sub(r'\s+', ' ', text.strip())


def get_file_mtime(path: Path) -> float:
    """Get file modification time."""
    return path.stat().st_mtime


def extract_snippet(text: str, query: str, max_length: int = 200) -> str:
    """Extract a snippet around the query in the text."""
    query_lower = query.lower()
    text_lower = text.lower()
    idx = text_lower.find(query_lower)
    if idx == -1:
        return text[:max_length] + "..." if len(text) > max_length else text
    start = max(0, idx - max_length // 2)
    end = min(len(text), idx + len(query) + max_length // 2)
    snippet = text[start:end]
    if start > 0:
        snippet = "..." + snippet
    if end < len(text):
        snippet += "..."
    return snippet


def walk_files(folder: Path, exts: List[str]) -> List[Path]:
    """Walk folder and return files with matching extensions."""
    files = []
    for ext in exts:
        files.extend(folder.rglob(f"*{ext}"))
    return sorted(set(files))  # dedupe and sort


def acquire_lock(lock_path: Path) -> bool:
    """Try to acquire a file lock. Returns True if acquired."""
    try:
        lock_path.touch(exist_ok=False)
        return True
    except FileExistsError:
        return False


def release_lock(lock_path: Path):
    """Release the file lock."""
    lock_path.unlink(missing_ok=True)


import argparse
import sys
from pathlib import Path
from .types import Config
from .server import MicroRAGServer


def main():
    import sys
    print("[DEBUG] Starting micro_rag_mcp server...", file=sys.stderr, flush=True)
    parser = argparse.ArgumentParser(description="Micro RAG MCP Server")
    parser.add_argument(
        "--data-folder",
        required=True,
        help="Path to the folder containing documents to index"
    )
    parser.add_argument(
        "--index-folder",
        help="Path to store the index (default: <data-folder>/index)"
    )
    parser.add_argument(
        "--exts",
        default=".txt,.md,.pdf,.docx",
        help="Comma-separated list of file extensions to index (default: .txt,.md,.pdf,.docx)"
    )
    parser.add_argument(
        "--rebuild-threshold",
        type=int,
        help="Optional threshold for full rebuild (not implemented yet)"
    )
    parser.add_argument(
        "--name",
        default="micro-rag-mcp",
        help="Name of the MCP server (default: micro-rag-mcp)"
    )
    parser.add_argument(
        "--description",
        help="Description of the MCP server for client applications"
    )

    args = parser.parse_args()

    print(f"[DEBUG] Parsed arguments, data_folder={args.data_folder}", file=sys.stderr, flush=True)
    
    data_folder = Path(args.data_folder).resolve()
    if not data_folder.is_dir():
        print(f"Error: {data_folder} is not a directory", file=sys.stderr)
        sys.exit(1)
    
    print(f"[DEBUG] Data folder validated: {data_folder}", file=sys.stderr, flush=True)

    index_folder = Path(args.index_folder) if args.index_folder else data_folder / "index"
    index_folder = index_folder.resolve()

    exts = [ext.strip() for ext in args.exts.split(",")]

    config = Config(
        data_folder=str(data_folder),
        index_folder=str(index_folder),
        exts=exts,
        rebuild_threshold=args.rebuild_threshold
    )

    print("[DEBUG] Creating MicroRAGServer instance...", file=sys.stderr, flush=True)
    server = MicroRAGServer(config, name=args.name, description=args.description)
    print("[DEBUG] MicroRAGServer created, starting server.run()...", file=sys.stderr, flush=True)
    server.run()


if __name__ == "__main__":
    main()


from pydantic import BaseModel
from typing import List, Optional


class ChunkRecord(BaseModel):
    id: int
    path: str
    ext: str
    mtime: float
    checksum: str
    chunk_index: int
    char_start: int
    char_end: int
    text: str
    embedding_dim: int
    deleted: bool = False


class FileRecord(BaseModel):
    path: str
    mtime: float
    checksum: str
    chunk_ids: List[int]
    ext: str


class SearchResult(BaseModel):
    path: str
    score: float
    snippet: str
    chunk_index: int
    char_start: int
    char_end: int
    mtime: float
    ext: str


class ReindexResult(BaseModel):
    files_added: int
    files_updated: int
    files_removed: int
    chunks_total: int
    elapsed_ms: int


class Config(BaseModel):
    data_folder: str
    index_folder: str
    exts: List[str]
    rebuild_threshold: Optional[int] = None


import json
import sys
import time
from pathlib import Path
from typing import Dict, List, Optional, Set
import faiss
import numpy as np
from .types import ChunkRecord, FileRecord, Config
from .embedder import Embedder
from .ingest import process_file
from .utils import get_file_mtime, walk_files, acquire_lock, release_lock, compute_checksum


class IndexManager:
    def __init__(self, config: Config, embedder: Embedder):
        print("[DEBUG] IndexManager.__init__ starting...", file=sys.stderr, flush=True)
        self.config = config
        self.embedder = embedder
        self.index_path = Path(config.index_folder) / "faiss.index"
        self.manifest_path = Path(config.index_folder) / "manifest.jsonl"
        self.files_path = Path(config.index_folder) / "files.json"
        self.lock_path = Path(config.index_folder) / "index.lock"
        self.version_path = Path(config.index_folder) / "version"
        self.index: Optional[faiss.Index] = None
        self.manifest: Dict[int, ChunkRecord] = {}
        self.files: Dict[str, FileRecord] = {}
        self.next_id = 0
        self.load_or_create()

    def load_or_create(self):
        """Load existing index or create new."""
        print("[DEBUG] IndexManager.load_or_create starting...", file=sys.stderr, flush=True)
        Path(self.config.index_folder).mkdir(parents=True, exist_ok=True)
        if self.index_path.exists() and self.manifest_path.exists():
            print("[DEBUG] Loading existing index...", file=sys.stderr, flush=True)
            self.load_index()
            print("[DEBUG] Index loaded successfully", file=sys.stderr, flush=True)
        else:
            print("[DEBUG] Creating new index (will initialize on first use)...", file=sys.stderr, flush=True)
            # Don't create index yet - will be created on first reindex/search
            self.index = None
            print("[DEBUG] Index initialization deferred", file=sys.stderr, flush=True)

    def load_index(self):
        """Load FAISS index and manifest."""
        self.index = faiss.read_index(str(self.index_path))
        with open(self.manifest_path, 'r') as f:
            for line in f:
                record = ChunkRecord(**json.loads(line))
                self.manifest[record.id] = record
                self.next_id = max(self.next_id, record.id + 1)
        if self.files_path.exists():
            with open(self.files_path, 'r') as f:
                data = json.load(f)
                self.files = {k: FileRecord(**v) for k, v in data.items()}

    def save_index(self):
        """Save FAISS index and manifest."""
        faiss.write_index(self.index, str(self.index_path))
        with open(self.manifest_path, 'w') as f:
            for record in self.manifest.values():
                f.write(record.model_dump_json() + '\n')
        with open(self.files_path, 'w') as f:
            json.dump({k: v.model_dump() for k, v in self.files.items()}, f, indent=2)
        self.version_path.write_text("1.0")

    def reindex(self, force: bool = False, path_glob: Optional[str] = None) -> dict:
        """Incremental reindex."""
        if not acquire_lock(self.lock_path):
            print("Reindex already in progress", file=sys.stderr)
            raise RuntimeError("Reindex already in progress")
        try:
            start_time = time.time()
            current_files = set()
            if path_glob:
                # Filter by glob
                all_files = walk_files(Path(self.config.data_folder), self.config.exts)
                current_files = {f for f in all_files if path_glob in str(f)}
                print(f"Reindexing with glob filter: {path_glob}, found {len(current_files)} files", file=sys.stderr)
            else:
                current_files = set(walk_files(Path(self.config.data_folder), self.config.exts))
                print(f"Reindexing all files, found {len(current_files)} files", file=sys.stderr)

            added, updated, removed = 0, 0, 0
            for file_path in current_files:
                rel_path = str(file_path.relative_to(Path(self.config.data_folder)))
                mtime = get_file_mtime(file_path)
                existing = self.files.get(rel_path)
                if existing and existing.mtime == mtime and not force:
                    continue
                chunks = process_file(file_path)
                if not chunks:
                    continue
                checksum = chunks[0][3]  # all chunks have same checksum
                if existing and existing.checksum == checksum and not force:
                    continue
                # Remove old chunks
                if existing:
                    for cid in existing.chunk_ids:
                        if cid in self.manifest:
                            self.manifest[cid].deleted = True
                    updated += 1
                else:
                    added += 1
                # Add new chunks
                chunk_ids = []
                texts = []
                for i, (text, start, end, _) in enumerate(chunks):
                    record = ChunkRecord(
                        id=self.next_id,
                        path=rel_path,
                        ext=file_path.suffix,
                        mtime=mtime,
                        checksum=checksum,
                        chunk_index=i,
                        char_start=start,
                        char_end=end,
                        text=text,
                        embedding_dim=self.embedder.dim
                    )
                    self.manifest[self.next_id] = record
                    chunk_ids.append(self.next_id)
                    texts.append(text)
                    self.next_id += 1
                # Initialize index if needed
                if self.index is None:
                    print("[DEBUG] Initializing index on first reindex...", file=sys.stderr, flush=True)
                    self.index = faiss.IndexFlatIP(self.embedder.dim)
                
                embeddings = self.embedder.encode_batch(texts)
                self.index.add(embeddings)
                self.files[rel_path] = FileRecord(
                    path=rel_path,
                    mtime=mtime,
                    checksum=checksum,
                    chunk_ids=chunk_ids,
                    ext=file_path.suffix
                )
            # Mark removed files
            for rel_path in list(self.files.keys()):
                if Path(self.config.data_folder) / rel_path not in current_files:
                    for cid in self.files[rel_path].chunk_ids:
                        if cid in self.manifest:
                            self.manifest[cid].deleted = True
                    del self.files[rel_path]
                    removed += 1
            self.save_index()
            elapsed = int((time.time() - start_time) * 1000)
            result = {
                "files_added": added,
                "files_updated": updated,
                "files_removed": removed,
                "chunks_total": len([r for r in self.manifest.values() if not r.deleted]),
                "elapsed_ms": elapsed
            }
            print(f"Reindex completed: {result}", file=sys.stderr)
            return result
        finally:
            release_lock(self.lock_path)

    def search(self, query: str, top_k: int = 5, score_threshold: float = 0.2) -> List[dict]:
        """Search the index."""
        # Initialize index if needed
        if self.index is None:
            print("[DEBUG] Initializing index on first search...", file=sys.stderr, flush=True)
            self.index = faiss.IndexFlatIP(self.embedder.dim)
            self.save_index()
        
        if self.index.ntotal == 0:
            print("Search: index empty", file=sys.stderr)
            return []
        query_emb = self.embedder.encode_single(query).reshape(1, -1)
        scores, indices = self.index.search(query_emb, min(top_k, self.index.ntotal))
        results = []
        for score, idx in zip(scores[0], indices[0]):
            if idx == -1 or score < score_threshold:
                continue
            record = self.manifest.get(idx)
            if record and not record.deleted:
                from .utils import extract_snippet
                snippet = extract_snippet(record.text, query)
                results.append({
                    "path": record.path,
                    "score": float(score),
                    "snippet": snippet,
                    "chunk_index": record.chunk_index,
                    "char_start": record.char_start,
                    "char_end": record.char_end,
                    "mtime": record.mtime,
                    "ext": record.ext
                })
        print(f"Search '{query}' returned {len(results)} results", file=sys.stderr)
        return results